@@ -57,16 +57,25 @@ class BaseST:
 
         return {"method": method, "url": url}
 
-    def _send_request(self, request, **kw):
+    def _send_request(self, request, dry=False, **kw):
         connection = self._connection
         func = getattr(self._session, request["method"])
-        return func(request["url"], auth=(connection["user"], connection["pwd"]), **kw)
+        if not dry:
+            return func(
+                request["url"], auth=(connection["user"], connection["pwd"]), **kw
+            )
+        else:
+            return
 
-    def _parse_response(self, request, resp):
+    def _parse_response(self, request, resp, dry=False):
         if request["method"] == "get":
             if resp.status_code == 200:
                 return resp.json()
         elif request["method"] == "post":
+
+            if dry:
+                return True
+
             if resp.status_code == 201:
                 m = IDREGEX.search(resp.headers.get("location", ""))
                 if m:
@@ -83,15 +92,18 @@ class BaseST:
         resp = self._parse_response(request, resp)
         if resp:
             return resp["value"]
+        else:
+            print("request failed", request)
 
-    def put(self):
+    def put(self, dry=False):
         if self._validate_payload():
             if self.exists():
                 return self.patch()
             else:
                 request = self._generate_request("post")
-                resp = self._send_request(request, json=self._payload)
-                return self._parse_response(request, resp)
+                resp = self._send_request(request, json=self._payload, dry=dry)
+
+                return self._parse_response(request, resp, dry=dry)
 
     def exists(self):
         name = self._payload["name"]
@@ -105,11 +117,11 @@ class BaseST:
             self.iotid = self._db_obj["@iot.id"]
             return True
 
-    def patch(self):
+    def patch(self, dry=False):
         if self._validate_payload():
             request = self._generate_request("patch")
-            resp = self._send_request(request, json=self._payload)
-            return self._parse_response(request, resp)
+            resp = self._send_request(request, json=self._payload, dry=dry)
+            return self._parse_response(request, resp, dry=dry)
 
 
 class Things(BaseST):
@@ -204,26 +216,134 @@ class Locations(BaseST):
     }
 
 
+class Sensors(BaseST):
+    _schema = {
+        "type": "object",
+        "required": ["name", "description", "encodingType", "metadata"],
+        "properties": {
+            "name": {"type": "string"},
+            "description": {"type": "string"},
+            "encodingType": {"type": "string"},
+            "metadata": {"type": "string"},
+        },
+    }
+
+
+class ObservedProperties(BaseST):
+    _schema = {
+        "type": "object",
+        "required": ["name", "definition", "description"],
+        "properties": {
+            "name": {"type": "string"},
+            "description": {"type": "string"},
+            "definition": {"type": "string"},
+        },
+    }
+
+
+class Datastreams(BaseST):
+    _schema = {
+        "type": "object",
+        "properties": {
+            "name": {"type": "string"},
+            "description": {"type": "string"},
+            "unitOfMeasurement": {
+                "type": "object",
+                "required": ["name", "symbol", "definition"],
+                "properties": {
+                    "name": {"type": "string"},
+                    "symbol": {"type": "string"},
+                    "definition": {"type": "string"},
+                },
+            },
+            "observationType": {"type": "string"},
+            "Thing": {
+                "type": "object",
+                "required": ["@iot.id"],
+                "properties": {"@iot.id": {"type": "number"}},
+            },
+            "ObservedProperty": {
+                "type": "object",
+                "required": ["@iot.id"],
+                "properties": {"@iot.id": {"type": "number"}},
+            },
+            "Sensor": {
+                "type": "object",
+                "required": ["@iot.id"],
+                "properties": {"@iot.id": {"type": "number"}},
+            },
+        },
+        "required": [
+            "name",
+            "description",
+            "unitofMeasurement",
+            "observationType",
+            "Thing",
+            "ObservedProperty",
+            "Sensor",
+        ],
+    }
+
+    def exists(self):
+        name = self._payload["name"]
+        thing = self._payload["Thing"]
+        lid = thing["@iot.id"]
+        resp = self.get(f"name eq '{name}'", entity=f"Things({lid})/Datastreams")
+
+        if resp:
+            try:
+                self._db_obj = resp[0]
+            except IndexError:
+                return
+
+            self.iotid = self._db_obj["@iot.id"]
+            return True
+
+
 class Client:
     def __init__(self, base_url, user, pwd):
         self._connection = {"base_url": base_url, "user": user, "pwd": pwd}
         self._session = Session()
 
-    def put_location(self, payload):
+    def put_sensor(self, payload, dry=False):
+        sensor = Sensors(payload, self._session, self._connection)
+        sensor.put(dry)
+        return sensor
+
+    def put_observed_property(self, payload, dry=False):
+        obs = ObservedProperties(payload, self._session, self._connection)
+        obs.put(dry)
+        return obs
+
+    def put_datastream(self, payload, dry=False):
+        datastream = Datastreams(payload, self._session, self._connection)
+        datastream.put(dry)
+        return datastream
+
+    def put_location(self, payload, dry=False):
         location = Locations(payload, self._session, self._connection)
-        location.put()
+        location.put(dry)
         return location
 
-    def put_thing(self, payload):
+    def put_thing(self, payload, dry=False):
         thing = Things(payload, self._session, self._connection)
-        thing.put()
+        thing.put(dry)
         return thing
 
-    def patch_location(self, iotid, payload):
+    def patch_location(self, iotid, payload, dry=False):
         location = Locations(payload, self._session, self._connection)
-        location.patch()
+        location.patch(dry)
         return location
 
+    def get_sensors(self, query=None):
+        yield from Sensors(None, self._session, self._connection).get(query)
+
+    def get_observed_properties(self, query=None):
+        yield from ObservedProperties(None, self._session, self._connection).get(query)
+
+    def get_datastreams(self, query=None):
+        yield from Datastreams(None, self._session, self._connection).get(query)
+
     def get_locations(self, query=None):
         yield from Locations(None, self._session, self._connection).get(query)
 
